

// g3ti_api.cpp : Defines the entry point for the console application.
//

#include <iostream>
#include "SangomaCard.h"

using namespace Sangoma;

#define MAX_HANDLES	256

const int VALID_NUM_OF_ARGS = 2;			///< The valid number of arguements expected by the program
const std::string A104_MODEL = "A104";		///< Command line string used to indicate to check the system for Sangoma A104 cards
const std::string A108_MODEL = "A108";		///< Command line string used to indicate to check the system for Sangoma A108 cards
bool gRunning = false;	///< Flag which is true if the sample program is currently monitoring the cards, false otherwise
	
std::vector<SangomaCard*> sangoma_cards;			//stores all the cards in the system
std::vector<SangomaPort*> sangoma_ports;			//stores ALL the ports in the system
std::vector<unsigned int> next_framing_type_index;	//for each port store the index of the next framing type to open the port with

#define DBG_MAIN	if(0)printf
#define INFO_MAIN	if(1)printf
#define WRN_MAIN	if(1)printf
#define ERR_MAIN	printf("Error: %s():line:%d: ", __FUNCTION__, __LINE__);printf

#define FUNC_MAIN()	if(1)printf("%s():line:%d\n", __FUNCTION__, __LINE__)

//SETUP FOR DETERMING PORT CONFIGURATION
const unsigned int NUMBER_OF_E1_FRAME_TYPES = 6;
const unsigned int NUMBER_OF_T1_FRAME_TYPES = 2;
const FramingType E1_TYPES[NUMBER_OF_E1_FRAME_TYPES] = {E1_CCS_CRC4_ON, E1_CCS_CRC4_OFF, E1_CAS_CRC4_ON, E1_CAS_CRC4_OFF, E1_CCS_UNFRAMED, E1_CAS_UNFRAMED};
const FramingType T1_TYPES[NUMBER_OF_T1_FRAME_TYPES] = {T1_EXTENDED_SUPERFRAME, T1_SUPERFRAME};	

static int iForcedLineType = -1; //< When greater than zero, will disable "Line Discovery" and configure for Line Type specified on command line.
static char szTxFileName[1024];	//< Full path to a file for transmission.

char* wp_get_time_stamp()
{
    static char formatted_time_stamp[400];

#if defined(__WINDOWS__)
    SYSTEMTIME timestamp;
    GetLocalTime(&timestamp);
	wp_snprintf(formatted_time_stamp, sizeof(formatted_time_stamp), "Y:%d/M:%d/D:%d. H:%d/M:%d/S:%d/MS:%u",
		timestamp.wYear, 
		timestamp.wMonth, 
		timestamp.wDay,
		timestamp.wHour, 
		timestamp.wMinute,
		timestamp.wSecond,
		timestamp.wMilliseconds);
#else
	struct timeval timestamp;
	gettimeofday(&timestamp, NULL);

	wp_snprintf(formatted_time_stamp, sizeof(formatted_time_stamp), "Sec:%u Ms:%u",
		(uint32_t)timestamp.tv_sec, (uint32_t)(timestamp.tv_usec / 1000));
#endif
	return formatted_time_stamp;
}   

void print_data_buffer(unsigned char *buffer, int buffer_length)
{
	printf("Data length: %u\n", buffer_length);

	for(int ln = 0; ln < buffer_length; ln++){
		if((ln % 20 == 0)){
			if(ln){
				printf("\n");
			}
			printf("%04d ", ln/20);//print offset
		}
		printf("%02X ", buffer[ln]);
	}
	printf("\n");
}

static int parse_command_line_args(
	int argc, char* argv[],
	CardModel *model,
	unsigned int *number_of_ports_per_card,
	Configuration & configuration,
	BufferSettings & buffer_settings)
{
	int i;
#define USAGE_STR \
"\n\
Usage: g3ti_api CardModel [LineType] [h MaxCableLossOption]\n\
\n\
Options:\n\
 CardModel             a104 or a108\n\
 LineType              T1 or E1\n\
 MaxCableLossOption    430, 360, 300, 225, 180, 175, 120\n\
 tx_tristate           this option will disable transmitter\n\
 -tx_file <TxFileName> Will continuesly transmit <TxFileName> when Line becomes synchronized.\n\
                       Must be used in conjuction with one of \"Line Type\" options because\n\
                       Line Tapping Device will discover the line type.\n\
 -e1_cas_crc4          Disable \"Line Discovery\" and configure for E1_CAS_CRC4_ON\n\
 -e1_cas_no_crc        Disable \"Line Discovery\" and configure for E1_CAS_CRC4_OFF\n\
 -e1_ccs_crc4          Disable \"Line Discovery\" and configure for E1_CCS_CRC4_ON\n\
 -e1_ccs_no_crc        Disable \"Line Discovery\" and configure for E1_CCS_CRC4_OFF\n\
 -e1_cas_unframed      Disable \"Line Discovery\" and configure for E1_CAS_UNFRAMED\n\
 -e1_ccs_unframed      Disable \"Line Discovery\" and configure for E1_CCS_UNFRAMED\n\
 -t1_esf               Disable \"Line Discovery\" and configure for T1_EXTENDED_SUPERFRAME\n\
 -t1_sf                Disable \"Line Discovery\" and configure for T1_SUPERFRAME\n\
\n\
Example 1: g3ti_api a108 e1\n\
Example 2: g3ti_api a108 t1\n\
Example 3: g3ti_api a108 e1 h 320\n\
Example 4: g3ti_api a108 e1 h 320 tx_tristate\n\
Example 4: g3ti_api a108 -e1_cas_crc4 -tx_file c:\\tmp\\test_file.pcm\n"

	*number_of_ports_per_card = 0;
	szTxFileName[0] = '\0';

	for(i = 0; i < argc; i++){
	
		if(wp_strncasecmp(argv[i], "?", strlen(argv[i])) == 0 || wp_strncasecmp(argv[i], "/?",  strlen(argv[i])) == 0){
			printf(USAGE_STR);
			return 1;
		}else if(wp_strncasecmp(argv[i], "T1",  strlen(argv[i])) == 0){
			configuration.e1Ort1 = T1;
		}else if(wp_strncasecmp(argv[i], "E1", strlen(argv[i])) == 0){
			configuration.e1Ort1 = E1;
		}else if(wp_strncasecmp(argv[i], "a104",  strlen(argv[i])) == 0){
			*model = A104;
			*number_of_ports_per_card = NUMBER_OF_A104_PORTS;
		}else if(wp_strncasecmp(argv[i], "a108",  strlen(argv[i])) == 0){
			*model = A108;
			*number_of_ports_per_card = NUMBER_OF_A108_PORTS;
		}else if(wp_strncasecmp(argv[i], "h",  strlen(argv[i])) == 0){
			if (i+1 > argc-1){
				ERR_MAIN("No Valid 'Max Cable Loss' value was provided!\n");
				printf(USAGE_STR);
				return 1;
			}
			switch(atoi(argv[i+1]))
			{
			case MCLV_43_0dB:
			case MCLV_36_0dB:
			case MCLV_30_0dB:
			case MCLV_22_5dB:
			case MCLV_18_0dB:
			case MCLV_17_5dB:
			case MCLV_12_0dB:
				configuration.HighImpedanceMode = true;
				configuration.MaxCableLoss = (MAX_CABLE_LOSS)atoi(argv[i+1]);
				break;
			default:
				configuration.HighImpedanceMode = false;
				configuration.MaxCableLoss = MCLV_12_0dB;
			}
		}else if(wp_strncasecmp(argv[i], "tx_tristate", strlen(argv[i])) == 0){
			configuration.TxTristateMode = true;
		
		}else if(wp_strncasecmp(argv[i], "bm",  strlen(argv[i])) == 0){
			if (i+1 > argc-1){
				ERR_MAIN("No Valid 'Buffer Multiplier' value was provided!\n");
				printf(USAGE_STR);
				return 1;
			}
			buffer_settings.BufferMultiplierFactor = atoi(argv[i+1]);
			if(	buffer_settings.BufferMultiplierFactor < MIN_BUFFER_MULTIPLIER_FACTOR ||
				buffer_settings.BufferMultiplierFactor > MAX_BUFFER_MULTIPLIER_FACTOR){
				ERR_MAIN("The 'Buffer Multiplier' value of %d is invalid! Minimum is: %d, maximum is: %d! Setting to %d.\n",
					buffer_settings.BufferMultiplierFactor, 
					MIN_BUFFER_MULTIPLIER_FACTOR, MAX_BUFFER_MULTIPLIER_FACTOR,
					MIN_BUFFER_MULTIPLIER_FACTOR);
				buffer_settings.BufferMultiplierFactor = MIN_BUFFER_MULTIPLIER_FACTOR;
			}else{
				INFO_MAIN("User's 'Buffer Multiplier' value is: %d.\n", buffer_settings.BufferMultiplierFactor);
			}

		}else if(wp_strncasecmp(argv[i], "-tx_file", strlen(argv[i])) == 0){
			if (i + 1 > argc - 1) {
				INFO_MAIN("No TxFileName provided!\n");
				printf(USAGE_STR);
				return 1;
			}
			wp_snprintf(szTxFileName, sizeof(szTxFileName), "%s", argv[i + 1]);
			INFO_MAIN("Setting szTxFileName to '%s'.\n", szTxFileName);

		}else if(wp_strncasecmp(argv[i], "-e1_cas_crc4", strlen(argv[i])) == 0){
			iForcedLineType = E1_CAS_CRC4_ON;
			configuration.e1Ort1 = E1;
		}else if(wp_strncasecmp(argv[i], "-e1_cas_no_crc", strlen(argv[i])) == 0){
			iForcedLineType = E1_CAS_CRC4_OFF;
			configuration.e1Ort1 = E1;
		}else if(wp_strncasecmp(argv[i], "-e1_ccs_crc4", strlen(argv[i])) == 0){
			iForcedLineType = E1_CCS_CRC4_ON;
			configuration.e1Ort1 = E1;
		}else if(wp_strncasecmp(argv[i], "-e1_ccs_no_crc", strlen(argv[i])) == 0){
			iForcedLineType = E1_CCS_CRC4_OFF;
			configuration.e1Ort1 = E1;
		}else if(wp_strncasecmp(argv[i], "-e1_cas_unframed", strlen(argv[i])) == 0){
			iForcedLineType = E1_CAS_UNFRAMED;
			configuration.e1Ort1 = E1;
		}else if(wp_strncasecmp(argv[i], "-e1_ccs_unframed", strlen(argv[i])) == 0){
			iForcedLineType = E1_CCS_UNFRAMED;
			configuration.e1Ort1 = E1;
		}else if(wp_strncasecmp(argv[i], "-t1_esf", strlen(argv[i])) == 0){
			iForcedLineType = T1_EXTENDED_SUPERFRAME;
			configuration.e1Ort1 = T1;
		}else if(wp_strncasecmp(argv[i], "-t1_sf", strlen(argv[i])) == 0){
			iForcedLineType = T1_SUPERFRAME;
			configuration.e1Ort1 = T1;
		}
	}//for()

	if(*number_of_ports_per_card == 0){
		printf(USAGE_STR);
		return 1;
	}

	if (szTxFileName[0] != '\0' && iForcedLineType == -1) {
		ERR_MAIN("\"Line Type\" must be provided for -tx_file option!\n");
		printf(USAGE_STR);
		return 1;
	}

	INFO_MAIN("\nCommand Line parsing summary:\n");

	if (iForcedLineType != -1) {
		configuration.Framing = (FramingType)iForcedLineType;
		buffer_settings.BufferMultiplierFactor = MIN_BUFFER_MULTIPLIER_FACTOR;
	}

	if (szTxFileName[0] != '\0') {
		//user wants to transmit a file - no need for Line Discovery
		INFO_MAIN("CardModel: %s, Forced Line Type: %s, TxFileName: '%s'\n\n", 
			(*model == A104 ? "A104":"A108"), DECODE_FRAMING(configuration.Framing), szTxFileName);

	} else {
		//user wants to tap a line - first must run Line Discovery
		INFO_MAIN("CardModel: %s, Line Type: %s, HighImpedance: %s, MaxCableLoss: %u, TxTristateMode: %s\n\n", 
			(*model == A104 ? "A104":"A108"), (configuration.e1Ort1 == T1 ? "T1":"E1"),
			(configuration.HighImpedanceMode == true ? "On":"Off"), configuration.MaxCableLoss,
			(configuration.TxTristateMode == true ? "On":"Off"));
	}
	INFO_MAIN("Press  <Enter> key to continue.\n");
	_getch();

	return 0;
}


///////////////////////////////////////////////////////////////////////////////////////////////
///	\fn		PortEventHandler(SangomaPort* p_port, unsigned int out_flags)
///	\brief	This function handles Data and Events raised by the Sangoma classes.
///	\param	p_port	Pointer to the SangomaPort object that issued this event
///	\param	out_flags	BitMap containing information about Data/Events raised by Sangoma Classes
///	\return	true if the event was successfully handled, false if an error occured
///	\author Josh Markwordt/David Rokhvarg
///	\date	5/31/2010
///////////////////////////////////////////////////////////////////////////////////////////////
bool PortEventHandler(SangomaPort* p_port, unsigned int out_flags)
{
	Alarms sangoma_alarms;
	//CHECK IF WE WERE PASSED A VALID PORT
	bool invalid_port = (NULL == p_port);
	if (invalid_port) {
		std::cout << "ERROR: PortEvent created on invalid port." << std::endl;
		return false;
	}
	
	if(p_port->GetIsOpen() == false){
		/* Some other port's event got signalled. Don't try to Read() on a closed port. */
		return true;
	}

	//READ FROM THE PORT THAT GENERATED THE EVENT (THIS READ NEEDS TO BE THREAD SAFE)
	//Regardless of the reason for the event always call Read().  This will either gives us the data or return
	//reason the event was triggered (e.g. PORT_IN_SYNC, PORT_OUT_OF_SYNC).
	int read_result = p_port->Read(out_flags);

	//PROCESS THE EVENT.
	//Also check if the read was successful, if not then we need to determine what the problem was.
	//DBG_MAIN("Port %u: ", p_port->GetPortNumber());
	switch(read_result)
	{
	case ReadDataAvailable:
		DBG_MAIN("ReadDataAvailable\n");
#if 0
		//As an example of how to access Rx data, print the Rx data:
		if (p_port->GetPortNumber() == 0) {
			print_data_buffer(p_port->GetRxDataBuffer(), p_port->GetRxDataLength());
		}
#endif
		//As an example of how to transmit data, call Write() whith Rx data as input.
//		p_port->Write(p_port->GetRxDataBuffer(), p_port->GetRxDataLength());
		break;
	case LineConnected:
		//DBG_MAIN("LineConnected\n");
		/// \todo Handle the line connected event
		//DAVIDR: port is open, line connected --> start handling incoming data.
		break;
	case LineDisconnected:
		//DBG_MAIN("LineDisconnected\n");
		//DAVIDR: the line got disconnected. If Alarams are of any interest, read them here.
		//May put a message for the user "cable disconnected".
		//
		sangoma_alarms = p_port->GetAlarms();
		/// \todo Handle the line disconnected event
		//DAVIDR Note: Some lines go up/down frequently and closing the port and starting Line type discovery
		//may cause a lot of unwanted load on the CPU. The best thing to do is to keep the port open
		//and wait for the LineConnected event, because the line type is already known and in the
		//real world it is not changing often. If the LineConnected event not received within
		//some timeout, it may make sense to close the port and start the Line type discovery.
		break;
	case ERROR_NO_DATA_AVILABLE:
		//FIXME: This condition may occur only during Line Discovery. Since no data can be received from
		//a disconnected port, this should be treated as a Warning, not an Error. Still, should be debugged in the future.
		//For now, ignore the warning.
		std::cout << "Port: " << p_port->GetPortNumber() << ": Warning: The event triggered but no data as available\n" << std::endl;
		break;
	case ERROR_INVALID_DESTINATION_BUFFER:
		std::cout << "Port: " << p_port->GetPortNumber() << ": Error: Unable to read because provided buffer was invalid\n" << std::endl;
		break;
	case ERROR_INVALID_DESTINATION_BUFFER_TOO_SMALL:
		std::cout << "Port: " << p_port->GetPortNumber() << ": Error: Unable to read because provided buffer was too small to hold the data.\n" << std::endl;
		break;
	case ERROR_PORT_NOT_OPEN:
		std::cout << "Port: " << p_port->GetPortNumber() << ": Error: The port is not Open.\n" << std::endl;
		break;
	case ERROR_DEVICE_IOCTL:
		std::cout << p_port->GetLastError();
		break;
	case TransmitBufferAvailable:
		p_port->WriteChunkOfTxFile();
		break;
	default:
		std::cout << "Port: " << p_port->GetPortNumber() << ": Unknown Error reported by Read()\n" << std::endl;
		break;
	} // switch(read_result)

	return true;
}
#if defined(__WINDOWS__)
DWORD SangomaThreadFunc(void *p_param)
#else
void* SangomaThreadFunc(void *p_param)
#endif
{ 
	int	iResult;
    DBG_MAIN("\n%s() - start\n", __FUNCTION__); 

	//CAST TO THE OBJECT THAT WAS PASSED IN
	std::vector<SangomaPort*>* p_ports = static_cast<std::vector<SangomaPort*>*>(p_param);

	SangomaPort	*p_port;

	//DETERMINE THE EVENT HANDLES FOR ALL THE PORTS IN ALL OF THE CARDS
	//these handles will be passed to WaitForMultipleObjects so that we know when to read
	unsigned int total_number_of_ports_in_system = static_cast<unsigned int>( p_ports->size() );
	unsigned int port_index, number_of_open_ports = 0;
	//SETUP VARIABLES FOR PROCESSING THE EVENT
	DWORD PORT_READ_TIMEOUT = 3000;		//arbitrarily set, doesn't matter b/c we don't do anything on a time out

	static sangoma_wait_obj* port_event_handles[MAX_HANDLES];
	static unsigned int port_input_flags[MAX_HANDLES];
	static unsigned int port_output_flags[MAX_HANDLES];
	static SangomaPort* sangoma_open_ports[MAX_HANDLES];	//stores Open ports only

	DBG_MAIN("%s():total_number_of_ports_in_system: %d\n", __FUNCTION__, total_number_of_ports_in_system); 

	//START PROCESSING IN THE THREAD
	while(gRunning)
	{
		//DavidR: Ports are getting Open and Closed during Line type Discovery, only
		//the Open ports will have a valid Event which can be waited on.
		//Go through the list of Open ports and populate the array of Sangoma Waitable objects.
		number_of_open_ports = 0;
		memset(port_event_handles, 0x00, sizeof(port_event_handles));
		memset(port_input_flags, 0x00, sizeof(port_input_flags));
		memset(port_output_flags, 0x00, sizeof(port_output_flags));
		memset(sangoma_open_ports, 0x00, sizeof(sangoma_open_ports));

		for(port_index = 0;
			port_index < total_number_of_ports_in_system && port_index < MAX_HANDLES;
			++port_index)
		{
			p_port = p_ports->at(port_index);

			if (p_port->GetIsOpen() == true &&
				p_port->GetEvent().EventHandle &&
				p_port->GetEvent().EventHandle != (void*)INVALID_HANDLE_VALUE) 
			{
				port_event_handles[number_of_open_ports] = p_port->GetEvent().EventHandle;

				if (szTxFileName[0] == '\0') {
					//rx and line state
					port_input_flags[number_of_open_ports] = (POLLIN | POLLPRI);
				} else {
					//tx and line state
					port_input_flags[number_of_open_ports] = (POLLOUT | POLLPRI);
				}
							
				sangoma_open_ports[number_of_open_ports] = p_port;

				number_of_open_ports++;
			}
		}

		if(!number_of_open_ports){
			WRN_MAIN("%s(): no open ports!\n", __FUNCTION__); 
			sangoma_msleep(1000);
			continue;
		}

		//DBG_MAIN("%s(): number_of_open_ports: %d\n", __FUNCTION__, number_of_open_ports); 

		//WAIT FOR AN EVENT FROM THE PORT
		//We are signalled anytime there is an event.  This could be for a line connection,
		//a line disconnection, or if there is data available.

		iResult = sangoma_waitfor_many(port_event_handles,
				port_input_flags,
				port_output_flags,
				number_of_open_ports,
				PORT_READ_TIMEOUT /* Wait timeout, in milliseconds. Or SANGOMA_INFINITE_API_POLL_WAIT. */
				);

		//DavidR: go through the array of OUT flags and do the work for those which are set.
		for(port_index = 0; port_index < number_of_open_ports; ++port_index)
		{
			switch(iResult)
			{
			case SANG_STATUS_APIPOLL_TIMEOUT:
				//timeout. try again.
				DBG_MAIN("%s(): Timeout\n", __FUNCTION__);
				continue;

			case SANG_STATUS_SUCCESS:
				/* a wait object was signaled */
				if(	port_output_flags[port_index] & POLLIN	||
					port_output_flags[port_index] & POLLPRI ||
					port_output_flags[port_index] & POLLOUT	)
				{
					// Retreive the signaled port from the list of open ports
					p_port = sangoma_open_ports[port_index];
					if(p_port == NULL){
						std::cout << "ERROR: invalid port index: " << port_index << std::endl;
					}
					
					// Call PortEventHandler() for each bit in port_output_flags[port_index]
					if (port_output_flags[port_index] & POLLIN) 
					{	// handle Rx Data
						if (PortEventHandler(p_port, POLLIN) == false) 
						{	// Call the event handler to take the appropriate actions
							std::cout << "ERROR: Rx Data on port " << port_index << " could not be handled properly." << std::endl;
						}
					}

					if (port_output_flags[port_index] & POLLPRI) 
					{	//handle an Event
						if (PortEventHandler(p_port, POLLPRI) == false) 
						{	// Call the event handler to take the appropriate actions
							std::cout << "ERROR: An Event on port " << port_index << " could not be handled properly." << std::endl;
						}
					}

					if (port_output_flags[port_index] & POLLOUT) 
					{	//handle an Event
						if (PortEventHandler(p_port, POLLOUT) == false) 
						{	// Call the event handler to take the appropriate actions
							std::cout << "ERROR: An Event on port " << port_index << " could not be handled properly." << std::endl;
						}
					}

				}

				if( port_output_flags[port_index] & (~port_input_flags[port_index]) ){
					WRN_MAIN("\nUnexpected port_output_flags[%d]: 0x%X\n\n", port_index, port_output_flags[port_index]);
					iResult = SANG_STATUS_IO_ERROR;
				}
				break;

			default:
				/* may get here when port being closed for re-configuration DURING THE WAIT. */
				WRN_MAIN("iResult: %s (%d)\n", SDLA_DECODE_SANG_STATUS(iResult), iResult);
			}//switch(iResult)

		}//for()

	} // while(gRunning)
	
    DBG_MAIN( "\n%s() - end\n", __FUNCTION__); 
	
#if defined(__WINDOWS__)
    return 0; 
#else
	pthread_exit(NULL);
    return NULL;
#endif
}

const char *decode_port_sync_status(SynchronizationStatus & status)
{
	switch( status )
	{
	case PORT_IN_SYNC: 
		return "PORT_IN_SYNC"; 
	case PORT_NOT_IN_SYNC:
		return "PORT_NOT_IN_SYNC"; 
	case PORT_WAITING_FOR_SYNC:
		return "PORT_WAITING_FOR_SYNC"; 
	}
	return "Unknown sync status";
}

void load_driver_modules()
{
#if defined(__WINDOWS__)
	//driver loaded by PnP system - do nothing here
#else
	//run "hwprobe" - if driver not loaded yet this command will load the driver
	system("wanrouter hwprobe");
#endif
}

void clear_screen()
{
#if defined(__WINDOWS__)
	system("cls");
#else
	system("clear");
#endif
}

void print_stats(void *p_param)
{
	std::vector<SangomaPort*>* p_ports = static_cast<std::vector<SangomaPort*>*>(p_param);

	clear_screen();
	
	INFO_MAIN("Time: %s\n", wp_get_time_stamp());
	//print port statistics/status
	for(unsigned int port_index = 0; port_index < p_ports->size(); ++port_index)
	{
		SangomaPort* p_port = p_ports->at(port_index);		//make a copy of the pointer to make it easier to work with
		SynchronizationStatus status = p_port->GetSynchronizationStatus();	//retrieve the synchronization status. DAVIDR: requires context switch to kernel-mode, should not be called too often.
		Statistics stats = p_port->GetStatistics();							//Get the statistics for this port

		INFO_MAIN("Port:%02d(unmapped:%02d):\tRx Count: %8u\tTx Count: %8u\n",
			p_port->GetPortNumber(), p_port->GetUnmappedPortNumber(),
			stats.BytesReceived, stats.BytesSent); 
	}
		
	INFO_MAIN("\n");

	//print Line Type of Ports in Sync
	for(unsigned int port_index = 0; port_index < p_ports->size(); ++port_index)
	{
		SangomaPort* p_port = p_ports->at(port_index);		//make a copy of the pointer to make it easier to work with
		SynchronizationStatus status = p_port->GetSynchronizationStatus();	//retrieve the synchronization status. DAVIDR: requires context switch to kernel-mode, should not be called too often.
		Configuration cfg = p_port->GetConfiguration();

		INFO_MAIN("Port %d: %20s,\tLineType: %s,\tFraming: %s\n",
			p_port->GetPortNumber(),
			decode_port_sync_status(status), 
			(cfg.e1Ort1 == E1 ? "E1" : "T1"), 
			DECODE_FRAMING(cfg.Framing));
	}
}

int main_program_loop(void * p_param, Configuration & configuration)
{
	//CAST TO THE OBJECT THAT WAS PASSED IN
	std::vector<SangomaPort*>* p_ports = static_cast<std::vector<SangomaPort*>*>(p_param);

	//CONTINUOUSLY TRY TO DETERMINE THE PORTS CONFIGURATION
	//if the port is in-sync nothing happens but while the port is out of sync
	//we cycle through the framing types to try and find the correct one.
	while( gRunning )
	{
		//DETERMINE EACH PORTS CONFIGURATION
		for(unsigned int port_index = 0; port_index < p_ports->size(); ++port_index)
		{
			//RETRIEVE THE SYNCHRONIZATION STATUS
			SangomaPort* p_port = p_ports->at(port_index);		//make a copy of the pointer to make it easier to work with
			SynchronizationStatus status = p_port->GetSynchronizationStatus();	//retrieve the synchronization status. DAVIDR: requires context switch to kernel-mode, should not be called too often.

			DBG_MAIN("Port %d: %s\n", port_index, decode_port_sync_status(status)); 
			//BASED ON THE STATUS DETERMINE WHAT TO DO
			switch( status )
			{
			case PORT_IN_SYNC: 
				//do nothing here, Read() will be performed in their separate threads
				break;
			case PORT_NOT_IN_SYNC:
				if (iForcedLineType != -1)
				{	//user specified the Line Type - do not attempt to run Line Discovery.
					if (p_port->GetIsOpen() == false) {
						configuration.lineCoding = ON;
						p_port->Open(configuration);//open the port with FORCED configuration, only a single time.

						if (szTxFileName[0] != '\0') {
							if (p_port->SetAndOpenTxFile(szTxFileName)) {
								ERR_MAIN("Failed to open file: %d\n", szTxFileName);
								exit(1);
							}
						}
					}
				} else {
					//determine the next configuration to try
					unsigned int next_frame_type = next_framing_type_index.at(port_index);						//retrieve the next frame type to try
					if(configuration.e1Ort1 == T1){
						next_framing_type_index.at(port_index) = (next_frame_type + 1) % NUMBER_OF_T1_FRAME_TYPES;	//move the stored index to the frame type to try next time
						configuration.Framing = T1_TYPES[next_frame_type];
						configuration.lineCoding = ON;
					}else{
						next_framing_type_index.at(port_index) = (next_frame_type + 1) % NUMBER_OF_E1_FRAME_TYPES;	//move the stored index to the frame type to try next time
						configuration.Framing = E1_TYPES[next_frame_type];
						configuration.lineCoding = ON;
					}

					p_port->Close();				//close the port
					p_port->Open(configuration);	//re-open the port with the new configuration
				}
				break;
			case PORT_WAITING_FOR_SYNC:
				//do nothing because we are waiting on a configuration to find sync
				break;
			default:
				//error, unknown synchronization status
				break;
			} // switch( status )
		} // for(unsigned int port_index = 0; port_index < ports.size(); ++port_index)

		//print some stats
		print_stats(&sangoma_ports);

		std::cout << "To refresh the screen press <Enter>" << std::endl;

#if defined(__WINDOWS__)
		std::cout << "To exit, press <x>" << std::endl;
#else
		std::cout << "To exit, press <x> and <Enter>" << std::endl;
#endif

		char key = _getch();
		switch(tolower(key))
		{
		case 'x':
		case 'q':
			gRunning = false;
			break;
		default:
			//std::cout << "Warning: Invalid command!" << std::endl;
			break;
		}
	
	} // while( gRunning )
	
	std::cout << "Exiting Main Loop ... " << std::endl;

	sangoma_msleep(2000);	//Sleep 2 seconds before exiting to make sure the other thread exits

	return 0;
}

int G3TI_CreateThread()
{
#if defined(__WINDOWS__)
	DWORD	SangomaThreadId;
	HANDLE	hSangomaThread;

	//step 1 - create the single worker thread
    hSangomaThread = CreateThread( 
        NULL,                       /* no security attributes        */ 
        0,                          /* use default stack size        */ 
        (LPTHREAD_START_ROUTINE)SangomaThreadFunc, /* thread function     */ 
        &sangoma_ports,				/* argument to thread function   */ 
        0,                          /* use default creation flags    */ 
        &SangomaThreadId			/* returns the thread identifier */ 
		);

	if(hSangomaThread == NULL){
		ERR_MAIN( "Failed to create Sangoma thread!!\n");
		return 1;
	}

#if 0
	//step 2 - set process priority to real-time
	if (!SetPriorityClass(GetCurrentProcess(), REALTIME_PRIORITY_CLASS)) {
		WRN_MAIN( "Failed to set Process priority!!\n");
	}

	//step 3 - set thread priority to real-time
	if(SetThreadPriority(hSangomaThread, THREAD_PRIORITY_TIME_CRITICAL) == FALSE){
		WRN_MAIN( "Failed to set Thread priority!!\n");
	}
#endif

	//success
	return 0;

#elif defined(__LINUX__)
# define CTHREAD_STACK_SIZE 1024 * 240
	pthread_t dwThreadId;
	pthread_attr_t attr;
	int result = -1;
	void *pParam = &sangoma_ports;

	result = pthread_attr_init(&attr);
   	//pthread_attr_setinheritsched(&attr, PTHREAD_EXPLICIT_SCHED);
	//pthread_attr_setschedpolicy(&attr, SCHED_RR);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
	pthread_attr_setstacksize(&attr, CTHREAD_STACK_SIZE);

	result = pthread_create(&dwThreadId, &attr, SangomaThreadFunc, pParam);
	pthread_attr_destroy(&attr);
	if (result) {
		return 1;
	}

	return 0;
#else
# error	Unsupported OS
#endif
}

int __cdecl main(int argc, char* argv[])
{
	unsigned int number_of_cards;
	struct BufferSettings buffer_settings;

	//create the configuration
	Configuration configuration;

	std::cout << "Application compiled on " << __DATE__ " " << __TIME__ << std::endl;
	std::cout << "Press  <Enter> key to continue." << std::endl;
	_getch();

	// PROCESS THE SUPPLIED INPUT
	// Parse the command line arguements to retreive the card model to check for.  At this
	// time, we only expect to use either A104 or A108s in the system, but not a mixture of the
	// two.
	CardModel model;
	unsigned int number_of_ports_per_card = 0;
	//default line type is T1
	configuration.e1Ort1 = T1;
	configuration.HighImpedanceMode = false;
	configuration.MaxCableLoss = MCLV_12_0dB;//default value for both Hi Impedance and Normal mode
	configuration.TxTristateMode = false;

	buffer_settings.BufferMultiplierFactor = MIN_BUFFER_MULTIPLIER_FACTOR;// go to higher mulitiplier only if CPU usage is high
	buffer_settings.NumberOfBuffersPerPort = MIN_NUMBER_OF_BUFFERS_PER_API_INTERFACE;

	if(parse_command_line_args(argc, argv, &model, &number_of_ports_per_card, configuration, buffer_settings)){
		return 1;
	}

	load_driver_modules();

	std::cout << "Search for '" << (model==A104?"A104":"A108") << "' Sangoma card." << std::endl;

	number_of_cards = GetNumberOfCardsInstalled(model);
	INFO_MAIN("Found %d %s card(s).\n", number_of_cards, (model==A104?"A104":"A108"));

	if(number_of_cards < 1){
		ERR_MAIN("No %s Sangoma cards running. Exiting...\n", (model==A104?"A104":"A108"));
		return 1;
	}

	for (unsigned int card_number = 0; card_number < number_of_cards; card_number++) 
	{
		SangomaCard* p_card = new SangomaCard(card_number, model, buffer_settings, buffer_settings);
		sangoma_cards.push_back(p_card);

		INFO_MAIN("\nCard Driver version: %i.%i.%i.%i, Firmware version: %s, PCI Slot: %i, PCI Bus: %i\n",
			p_card->GetCardDriverVersion().Major, p_card->GetCardDriverVersion().Minor,
			p_card->GetCardDriverVersion().Build, p_card->GetCardDriverVersion().Revision,
			p_card->GetFirmwareVersion().c_str(), p_card->GetCardPciBus(), p_card->GetCardPciSlot());
	
		//Add this cards ports to the vector containing all the system's ports
		unsigned int number_of_ports = p_card->GetNumberOfPorts();
		for(unsigned int port_index = 0; port_index < number_of_ports; ++port_index)
		{
			sangoma_ports.push_back( p_card->GetPort(port_index) );
			next_framing_type_index.push_back(0);	//Start with the first framing type, this indexes T1/E1_TYPES to find
													//the frame type to use.
		}
	} // for (unsigned int card_number = 0; card_number < number_of_cards; card_number++)

	std::cout << "Press  <Enter> key to continue." << std::endl;
	_getch();

	gRunning = true;
	if ( G3TI_CreateThread())
	{
		ERR_MAIN("Failed to create receiver thread!\n");
		return 1;
	}

	main_program_loop(&sangoma_ports, configuration);

	//DELETE ALL THE SANGOMA CARDS
	for(unsigned int card_index = 0; card_index < sangoma_cards.size(); ++card_index)
	{
		delete sangoma_cards.at(card_index);
	}

	INFO_MAIN("\n%s: Done.\n", argv[0]);
	return 0;
}


